// Visites_Appel_AVH_v2     
// Prog d'enregistrement et de téléchargement des visites et des appels à l'AVH35
// Compiler avec ESP32 Dev Module (ESP32 Whadda) ou DOIT ESP32 DEVKIT v1 (ESP32 DEVKIT v4)
// Matériel  = ESP32 38 pins + SDcard SPI 3,3v. + module RTC DS2405 I2C 3,3v. + Buzzer + 2 boutons + 2 condens 100uF 
//           + 2 leds jaune + 1 led orange + 3 resist. 220 ohms
// Un bouton sert à enregistrer l'heure des visites, l'autre bouton sert à enregistrer l'heure des appels
// Principe fonctionnement =
//  A chaque appui sur l'un des boutons : enregistrement de l'appel ou de la visite avec horodatage dans le fichier du jour.
//  A chaque appui (pour confirmer la saisie) :
//     >>> bouton visite : le buzzer imite le bruit d'un carillon de porte 
//                         la led associée au bouton visite s'allume pendant 2 sec.
//     >>> bouton appel : le buzzer imite le bruit d'une sonnerie de téléphone.
//                        la led associée au bouton appel s'allume pendant 2 sec.  
//  En cas d'erreur :
//     >>> si erreur de connexion wifi : le buzzer imite une sirène de pompiers
//                                       la Led orange reste ensuite allumée en permanence  (erreur non bloquante)                           
//     >>> si erreur sur la carte SD : le buzzer imite une sirène d'ambulance 
//                                     la Led orange clignote continuellement (erreur bloquante)
//  A la demande : consultation en Wifi en tapant l'ip. du module enregistreur + le n° de port dans un navigateur internet d'un PC 
//     >>> La page HTML d'accueil affiche le décompte des données saisies mois par mois sur les 12 derniers mois
//     >>> Un premier bouton virtuel  permet de télécharger les données du mois précédent 
//     >>> Un second bouton virtuel  permet de télécharger les données de l'un des 12 derniers mois (à choisir)  
//     >>> Un troisième bouton virtuel  permet de télécharger les données pour une période donnée (à définir)
//     >>> Le fichier demandé est  téléchargé au format CSV dans le PC
//     >>> L'historique des saisies est toujours conservé sur la carte SD.
// Pour plus de détails, voir la documentation relative à cette application.
// Réalisation : Yves Le Chevalier/MHK pour AVH-35  (Juin 2024)

// Brochage :
//  pin 22  ==>  SCL mod.RTC (I2C)
//  pin 21  ==>  SDA mod.RTC (I2C)
//  pin 16  ==>  bouton 1 (pullup)   bouton appel
//  pin 17  ==>  bouton 2 (pullup)   bouton visite
//  pin 15    ==>  buzzer
//  pin 18  ==>  SCK carte SD (SPI)
//  pin 19  ==>  MISO carte SD (SPI)
//  pin 23  ==>  MOSI carte SD (SPI)
//  pin 5  ==>  SS carte SD (SPI)
//  pin 32  ==> led témoin appui bouton 1 (appel)
//  pin 33  ==> led témoin appui bouton 2 (visite)
//  pin 34  ==> led témoin erreur SD ou Wifi

//------------------------------------------------------------------
#define TRACE 0   // mise au point : 1 pour activer trace sur port série, 0 pour désactiver
//------------------------------------------------------------------
// Prototypage des fonctions
void maj_rtc();
void afftablecumuls();
void calcperiodeaff();
void blocsaisie();
void deblocsaisie();
void trmt_appel();
void trmt_visite();
//String get_wifi_status(int status);
void erreur_SD();
void erreur_wifi();
void beep_depart();
void beep_visite();
void beep_appel();
void initsdcard();
void ouvhistoecr();
void enrsaisie();
void closhisto();
void creselectecr();
void ouvseleclec();
void closelect();
void effselec();
void ouvhistolec();
void ouvarchivecr();
void cloarchiv();
void renomhisto();
void renomselect();
void comptasaisie();
void lecparam();
void pagerreur404();
void pageaccueil();
void handle_NotFound();
void accueil();
void transfertmois();
void trmoisprec();
void uploadselec();
void recuhisto();
void pageperiod();
void demandperiod();
void selechisto();
void chargehisto();
void selectmois();
void trautremois();
void pagefinarch();
void archivage();

//------------------------------------------------------------------
#include <Arduino.h>
#include <WiFi.h>                  // WiFi version 2.0.0
#include <WebServer.h>             // WebServer version 2.0.0
int port = 1081;                   // port de routage
WebServer server(port);            // on instancie un serveur sur le port 1081 
char ssid[50]      = "";           // nom du réseau wifi
char password[100] = "";           // mot de passe réseau wifi
unsigned long timout = 0L;         // timbre de calcul du timeout de la connexion Wifi
String page = "";                  // chaine pour contenir la page à afficher
bool errwf = false;                // indicateur d'erreur connexion Wifi
bool finconnex = false;            // indicateur fin de boucle de connexion wifi
String hostname = "ESP32_Visites_Appels_AVH";

//------------ Buzzer   
#define Buzz_pin 15                // buzzer sur pin 4
#define Buzz_channel 1             // canal PWM 1

//------------ Boutons   
#define Bout1   2                  // bouton 1 => enregistrement des appels
#define Bout2   4                  // bouton 2 => enregistrement des visites 
bool appel = false;                // indicateur d'appui sur le bouton appel
bool visite = false;               // indicateur d'appui sur le bouton visite
unsigned long timbou1,timbou2 = 0; // variables de temporisation pour éviter le rebond sur appui bouton
char timestamp[20] = "";           // format d'horodatage
bool blocage = false;              // indicateur de blocage de la saisie pendant un transfert

//------------ Leds   
#define LEDVIS   32                // Led témoin d'enregistrement d'une visite
#define LEDAPP   33                // Led témoin d'enregistrement d'un appel
#define LEDERR   25                // Led témoin d'une erreur SD ou Wifi

//------------ Serveur NTP  
const char *ntpServer = "ntp.obspm.fr";  // Serveur NTP observatoire de Paris
String an4 = "0000";
String an2 = "00";
String datheur = "";
int an = 0;
int mois = 0;
int jour = 0; 
int heur = 0;
int minut = 0;
int second = 0;
int numjou = 0;
unsigned long timntp = 0;
bool ntpok = false;

//------------ RTC 
#include "GravityRtc.h"            // GravityRtc version 1.0.1  (pour RTC DS2405)
#include "Wire.h"                  // Wire version 2.0.0
GravityRtc rtc;     
char updatim[20] = "";             // RTC updated time

//------------ Carte SD 
#include "SD.h"                    // SD version 2.0.0  +  FS version 2.0.0
#include "SPI.h"                   // SPI version 2.0.0
#define SDSSPIN 5                  // SD card CS/SS pin 5
bool sdok = true;                  // indicateur erreur/ok du traitmt. carte SD
File fichier1;                     // objet fichier  
File fichier4;                     // objet fichier  
File fichier5;                     // objet fichier  
File fichier6;                     // objet fichier  
File fichier7;                     // objet fichier  
char nomfich3[11] = "/Haamm.txt";  // nom ancien historique apres archivage 
char nomfich4[11] = "/ARCHI.txt";  // nom du fichier archives
char nomfich5[11] = "/PARAM.txt";  // nom du fichier avec identif wifi
char nomfich6[11] = "/HISTO.txt";  // nom du fichier historique des saisies
char nomfich7[11] = "/SELEC.txt";  // nom du fichierde selection
String nomtransf = "";
char datew[12];                    // formatage de la date
char typenr;                       // identifiant visite ou appel
int cptenr = 0;                    // compteur d'enregistrements 
int cptper = 0;                    // compteur d'enregistrements de la péruode traitée
int cptold = 0;                    // compteur d'enregistrements antérieurs à la période traitée
int cpthis = 0;
int cptcons = 0;

//----------teléchargement historique
String dadeb = "";           // date de début sélection histo 
String dafin = "";           // date de fin sélection histo 
String debsel = "";          // date de début sélection histo demandée AAMMJJ
String finsel = "";          // date de fin sélection histo demandée AAMMJJ
String dselaff = "";         // date de début sélection histo demandée jjmmaaaa
String fselaff = "";         // date de fin sélection histo demandée jjmmaaaa
String dselfic = "";         // date de début sélection pour nom fichier jjmmaa 
String fselfic = "";         // date de fin sélection pour nom fichier jjmmaa 

//---------Variables diverses
int i,j = 0;
int moisec = 0;             // mois en cours
int anec = 0;               // année en cours
int anprec = 0;             // année précédente
int tmois[13];              // mois des 12 derniers mois précédent le mois en cours, [12] = mois en cours
int tans[13];               // ans des 12 derniers mois précédent le mois en cours, [12] = mois en cours
int nbvis[13];              // cummul des visites pour chaque mois 
int nbapp[13];              // cummul des appels pour chaque mois 
char *nommois[] = {"Janvier","Février","Mars","Avril","Mai","Juin","Juillet","Aout","Septembre","Octobre","Novembre","Décembre"}; 
bool inselmois = false;
int moisselec;
int indmois;

//------------------------------------------------------------------
void maj_rtc() {           // récup date et heure sur serveur NTP pour mise a jour RTC
  configTzTime("CET-1CEST-2,M3.5.0/02:00:00,M10.5.0/03:00:00", ntpServer);   // changement auto heure été/hiver
  delay (2000);
  char buffer[30];
  unsigned long timntp = millis();
  while (an2 != "20") {                      // attente de recevoir la date/heure du moment
    time_t timestamp = time(NULL);
    struct tm *pTime = localtime(&timestamp);
    strftime(buffer, 30, "%d/%m/%Y %H:%M:%S %u", pTime);
    datheur = String(buffer);
    an2 = datheur.substring(6,8);
    an4 = datheur.substring(6,10);
    delay (3000);                              // attente 3 sec avant de retenter la connexion NTP
    if (millis() - timntp > 12000)  an2 = "20";   // au bout de 12 sec. abandon de l'attente et sortie de la boucle
  }
  if (an4 != "0000") {                         // si année modifiée => date et heure reçues du serveur NTP
    an = (datheur.substring(6,10)).toInt();
    mois = (datheur.substring(3,5)).toInt();
    jour = (datheur.substring(0,2)).toInt();
    heur = (datheur.substring(11,13)).toInt();
    minut = (datheur.substring(14,16)).toInt();
    second = (datheur.substring(17,19)).toInt();
    numjou = (datheur.substring(20,21)).toInt();    // N° de jour dans la semaine (1 = lundi)
    rtc.adjustRtc(an,mois,jour,numjou,heur,minut,second);        // mise a jour du RTC
    ntpok = true;                 // indique maj rtc par le NTP effectuée
    sprintf(updatim, "%04d%02d%02d;%02d%02d%02d;%01d", an, mois, jour, heur, minut, second, numjou);
    #if TRACE
      Serial.print("RTC mis a jour avec : ");
      Serial.println(String(updatim));  
      rtc.read(); 
      Serial.print("Lect.RTC (contrôle) = "); 
      Serial.print(rtc.year); 
      Serial.print("  "); 
      Serial.print(rtc.month); 
      Serial.print("  "); 
      Serial.print(rtc.day); 
      Serial.print("  "); 
      Serial.print(rtc.hour); 
      Serial.print("  "); 
      Serial.print(rtc.minute); 
      Serial.print("  "); 
      Serial.print(rtc.second); 
      Serial.print("  ");  
      Serial.print(rtc.week); 
      Serial.println("  "); 
    #endif   
  }
  else ntpok = false;         // indique maj rtc par le NTP a échoué
}

//------------------------------------------------------------------
void afftablecumuls() {     // affichage de contrôle des comptages sur la console
  #if TRACE
    for (i = 0; i < 13; i++) {
      Serial.print(tmois[i]); 
      Serial.print(" ");
      Serial.print(tans[i]); 
      Serial.print(" : ");
      Serial.print(nbvis[i]); 
      Serial.print(" ");
      Serial.print(nbapp[i]);
      Serial.print(" / "); 
    }
    Serial.println(""); 
  #endif 
}

//------------------------------------------------------------------
void calcperiodeaff() {       // calcul des 12 mois précédent
  tmois[12] = moisec; 
  tans[12] = anec;
  for (i = 11; i > -1; i--) {
    j = i + 1;
    tmois[i] = tmois[j] - 1;
    tans[i] = tans[j];
    if (tmois[i] == 0) {
      tans[i]--;
      tmois[i] = 12; 
    }
    nbvis[i] = 0;           // et RAZ des comptages par mois
    nbapp[i] = 0;
  }
  #if TRACE
    afftablecumuls();
  #endif 
}

//------------------------------------------------------------------
void blocsaisie() {           // blocage saisie pendant opération de transfert
  blocage = true;
  digitalWrite(LEDVIS,HIGH);
  digitalWrite(LEDAPP,HIGH);
  ledcWriteTone(Buzz_channel,75);  
}

//------------------------------------------------------------------
void deblocsaisie() {           // déblocage saisie après opération de transfert
  blocage = false;
  digitalWrite(LEDVIS,LOW);
  digitalWrite(LEDAPP,LOW);
  ledcWriteTone(Buzz_channel,0);  
}

//------------------------------------------------------------------
void IRAM_ATTR trmt_appel() {
  if (!blocage) {
    if (millis() - timbou1 > 1000) {   // temporisation 1 sec anti-rebond du bouton
      appel = true;
      timbou1 = millis();              // mémorisation de l'heure d'appui sur le bouton 
    }
  }
}

//------------------------------------------------------------------
void IRAM_ATTR trmt_visite() {
  if (!blocage) {
    if (millis() - timbou2 > 1000) {   // temporisation 1 sec anti-rebond du bouton
      visite = true;
      timbou2 = millis();              // mémorisation de l'heure d'appui sur le bouton 
    }
  }
}

//------------------------------------------------------------------
String get_wifi_status(int status) {
  switch(status) {               // codes retour de connexion au wifi
    case WL_IDLE_STATUS:
      return "WL_IDLE_STATUS";
    case WL_SCAN_COMPLETED:
      return "WL_SCAN_COMPLETED";
    case WL_NO_SSID_AVAIL:
      return "WL_NO_SSID_AVAIL";
    case WL_CONNECT_FAILED:
      return "WL_CONNECT_FAILED";
    case WL_CONNECTION_LOST:
      return "WL_CONNECTION_LOST";
    case WL_CONNECTED:
      return "WL_CONNECTED";
    case WL_DISCONNECTED:
      return "WL_DISCONNECTED";
    other:
      return "?????";
  }
}
 
//------------------------------------------------------------------
void setup() {  
  Serial.begin(115200);                 // port série pour controle
  pinMode(Bout1, INPUT_PULLUP);
  pinMode(Bout2, INPUT_PULLUP);
  pinMode(LEDVIS, OUTPUT);
  pinMode(LEDAPP, OUTPUT);
  pinMode(LEDERR, OUTPUT);
  digitalWrite(LEDERR,LOW);
  blocsaisie();

  rtc.setup();                         // start DS2405 RTC
  
  Serial.println("");
  Serial.println("Visites et appels AVH");
  Serial.println("");

  ledcSetup(Buzz_channel, 2000, 8);      // réglage fréquence et résolution du canal
  ledcAttachPin(Buzz_pin, Buzz_channel); // attachement buzzer sur canal 0
  ledcWriteTone(Buzz_channel,0);         // coupe le son

  initsdcard();                          // initialisation de la carte SD
  if (!sdok) {                           // Carte SD absente
    Serial.println("Erreur presence carte SD ==> Arrêt programme"); 
    erreur_SD();
  } 
  delay(500);
  lecparam();           // test si présence de fichier paramètres pour modifier SSID et PWD par défaut

  WiFi.disconnect();
  int status = WL_IDLE_STATUS;
  timout = millis(); 
  #if TRACE
    Serial.print ("ssid=");
    Serial.println(ssid);
    Serial.print ("pwd=");
    Serial.println(password);
  #endif
  WiFi.config(INADDR_NONE, INADDR_NONE, INADDR_NONE, INADDR_NONE);
  WiFi.mode(WIFI_STA);
  WiFi.setHostname(hostname.c_str());
  WiFi.begin(ssid, password);      // connexion au Wifi du réseau       
  do {
    if (WiFi.status() == WL_CONNECTED)  finconnex = true;    // pour sortie de boucle de connexion
    else {
      #if TRACE
        status = WiFi.status();
        Serial.print("Wifi status = "); 
        Serial.println(get_wifi_status(status));
      #endif 
      delay(450);
      #if TRACE
        Serial.print(".");       // barre de progression de la connexion wifi
      #endif  
      ledcWriteTone(Buzz_channel,1000);      
      delay(50);
      ledcWriteTone(Buzz_channel,0); 
      if (millis() - timout > 30000) { // timout connexion wifi au bout de 30 sec
        #if TRACE
          Serial.println("Station non connectée au réseau Wifi");
        #endif  
        errwf = true;                  // indic. erreur wifi
        finconnex = true;              // pour sortie de boucle de connexion
        erreur_wifi();
        digitalWrite(LEDERR,HIGH);     // led témoin erreur reste allumée
        delay(1000);
      }
    }
  }   
  while (!finconnex); 

  if (!errwf) {                               // si connexion wifi ok
    WiFi.setAutoReconnect(true);
    WiFi.persistent(true);
    // affiché sytématiquement pour avoir l'adresse IP de connexion
    Serial.println("");               
    Serial.print("Station mode  SSID : ");  // réseau local pour connexion au wifi
    Serial.print(ssid);
    Serial.print("  IP address = ");     // adresse ip pour accéder au module en wifi
    Serial.print(WiFi.localIP());
    Serial.print(":");
    Serial.print(port);                  // n° de port pour accéder au module en wifi
    Serial.println(""); 

    maj_rtc();                             // réglage date et heure à partir date et heure du serveur NTP
  
    server.begin();                        // démarrage du serveur 
    Serial.println("Serveur HTTP demarre");        
    server.on("/", accueil);               // affichage de la page d'accueil 
    server.on("/trmois", trmoisprec);          // télécharger la saisie du mois précédent
    server.on("/trautre", trautremois);    // saisir le mois a télécharger
    server.on("/trperiod", demandperiod);  // saisir date début et fin de période à sélecter dans historique
    server.on("/submit", selechisto);      // extraitre les données pour la période saisie
    server.on("/charghis", chargehisto);   // téléchager le fichier de sélection dans historique
    server.on("/archiv", archivage);   // téléchager le fichier de sélection dans historique
    server.onNotFound(handle_NotFound);    // traitement d'une demande inconnue
  }

  attachInterrupt(Bout1, trmt_appel, FALLING);    // mise à la masse quand appui sur un bouton
  attachInterrupt(Bout2, trmt_visite, FALLING); 

  rtc.read(); 
  anec = rtc.year;
  moisec = rtc.month;
  #if TRACE
    Serial.print("Mois/année en cours : ");  
    Serial.print(moisec); 
    Serial.print(" ");
    Serial.println(anec); 
  #endif
  calcperiodeaff();
  deblocsaisie();
  beep_depart();                   // son arpège pour indiquer que le système est opérationnel
}

//------------------------------------------------------------------
void loop()  {
  if (appel)  {
    digitalWrite(LEDAPP,HIGH);    // led témoin appel allumée
    appel = false;
    beep_appel();
    typenr = 'A';
    enrsaisie();
    digitalWrite(LEDAPP,LOW);     // led témoin appel éteinte
  }
  if (visite)  {
    digitalWrite(LEDVIS,HIGH);    // led témoin visite allumée
    visite = false;
    beep_visite();
    typenr = 'V';
    enrsaisie();
    digitalWrite(LEDVIS,LOW);     // led témoin visite éteinte
  }
  if (!errwf) {                   // si connexion wifi ok
    // à chaque itération, on appelle handleClient pour que les requêtes Wifi soient traitées
    server.handleClient();
    yield();
  }
}

//________________________________End_______________________________
